# See end of file for licensing information

package AppLib::HTML;

# The AppLib::HTML package holds commonly used HTML related methods, such 
# as table generation and trees etc.

# Sections in this module are listed in bottom up order.

use strict;
use UNIVERSAL 'isa';
use base 'AppLib::Error';





#####################################################################
# Escaping Methods

# First, provide some resources.
use vars qw{%escapeMap $text $property $textarea};
BEGIN {
	%escapeMap = (
		'&' => '&amp;',
		'<' => '&lt;',
		'>' => '&gt;',
		'"' => '&quot;',
		"'" => '&#039;',
		"\n" => "<br>\n",
		);
	$text = qr/([&<>])/;
	$property = qr/([&<>"'])/;
	$textarea = qr/([&<>"'])/;
}

# Escape "text" in general
sub escapeText { $_[1] =~ s/([&<>])/$escapeMap{$1}/g; $_[1] }

# Escape for use in a tag property
sub escapeProperty { $_[1] =~ s/([&<>"'])/$escapeMap{$1}/g; $_[1] }

# Escape for use in a textarea
sub escapeTextArea { $_[1] =~ s/([&<>"'])/$escapeMap{$1}/g; $_[1] }

# Escape text to be displayed on an html page
sub escapeHTMLText {
	$_[1] =~ s/([&<>])/$escapeMap{$1}/g;
	$_[1] =~ s/(\015\012|\015|\012)/<br>\n/g;
	return $_[1];
}





#####################################################################
# Callback Generators

# Create a generic callback link.
# Don't escape, that is left to the caller
sub linkCallback {
	my $class = shift;
	my $text = shift;
	my $args = shift || {};
	my $options = shift || {};
	return undef unless isa( $args, 'HASH' );
	return undef unless isa( $options, 'HASH' );
	
	# Copy the options and args to avoid 
	# corrupting passed values
	$args = { %$args };
	$options = { %$options };
	
	# Avoid warnings in the map below
	foreach ( keys %$args ) {
		$args->{$_} = '' unless defined $args->{$_};
	}
	
	# Create the href
	my $arg_string = join( '&', map { "$_=$args->{$_}" } keys %$args );
	my $href = length($arg_string)
		? "$ENV{SCRIPT_NAME}?$arg_string"
		: "$ENV{SCRIPT_NAME}";
	
	# Create the full set of tag options
	$options->{href} = $href;
	return $class->tagPair( 'a', $options, $text );
}	





#####################################################################
# String generators

# Generates an arbitrary tag, somewhat like CGI.pm does.
# To reduce load, this method does not do escaping or case alteration.
# Escaping untrusted strings is left to the caller.
sub tag {
	my $class = shift;
	my $tag = shift or return undef;
	my $props = shift or return undef;
	return '<' 
		. join(' ', $tag, map { defined $props->{$_} 
			? "$_=\"$props->{$_}\""
			: "$_"
			} keys %$props) 
		. '/>';
}

# Generate a pair of tags with content in them.
# To reduce load, this method does not do escaping or case alteration.
# Escaping untrusted strings is left to the caller.
sub tagPair {
	my $class = shift;
	my $tag = shift or return undef;
	my $props = shift or return undef;
		
	# Generate the open tag ( as above )
	my $html = '<' 
		. join(' ', $tag, map { defined $props->{$_} 
			? "$_=\"$props->{$_}\""
			: "$_"
			} keys %$props) 
		. '>';
	
	# Add the content ( escaping is left to the caller )
	# and the closing tag, and return.
	return $html . join( '', @_ ) . "</$tag>";
}

1;

__END__

# Copyright (C) 2000-2002 Adam Kennedy ( software.applib@ali.as )
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# Should you wish to utilise this software under a different licence, 
# please contact the author.

