# See end of file for licensing information

package AppLib::HTML::Table;

# Package to generate, large, complex tables.

use strict;
use base 'AppLib::Error';






# Generate a complete formatted table.
use vars qw{$defaultcolor};
BEGIN {
	$defaultcolor = {
		bg => '#FFFFFF',
		title => '#CCCCCC',
		total => '#EEEEEE',
		border => '#000000',
		altrow => '',
		};
}

sub setDefaultColor {
	my ( $class, $name, $value ) = @_;
	return undef unless exists $defaultcolor->{$name};
	$defaultcolor->{$name} = $value;
	return 1;
}		

sub table {
	my $class = shift;
	my $data = shift;
	my $opt = shift;

	# Get the rows section
	my $content = $class->tableRows( $data, $opt );

	# Wrap and return
	return $class->tableWrap( $content, $opt );
}

# Generate just the rows portion of the table, without the wrapping
sub tableRows {
	my $class = shift;
	my $data = shift;
	my $opt = shift;

	# Set colors
	$opt->{bgcolor} = $defaultcolor->{bg} unless $opt->{bgcolor};
	$opt->{titlecolor} = $defaultcolor->{title} unless $opt->{titlecolor};
	$opt->{totalcolor} = $defaultcolor->{total} unless $opt->{totalcolor};
	if ( $defaultcolor->{altrowcolor} and ! $opt->{altrowcolor} ) {
		$opt->{altrowcolor} = $defaultcolor->{altrow};
	}

	# Get the depth and width of the table
	$opt->{rows} = $#$data;
	$opt->{columns} = 0;
	foreach ( map { $#$_ } @$data ) {
		$opt->{columns} = $_ if $_ > $opt->{columns};
	}
		
	# Create the properties map
	my %properties = (
		top => {
			color => ($opt->{topcolor} || $opt->{titlecolor}),
			align => ($opt->{topalign} || 'center'),
			bold  => (exists $opt->{boldtop} ? $opt->{boldtop} : 1),
			class => $opt->{topclass},
			},
		bottom => {
			color => ($opt->{bottomcolor} || $opt->{totalcolor}),
			align => ($opt->{bottomalign} || 'right'),
			bold  => (exists $opt->{boldbottom} ? $opt->{boldbottom} : 1),
			class => $opt->{bottomclass},
			},
		left => {
			color => ($opt->{leftcolor} || $opt->{titlecolor}),
			align => ($opt->{leftalign} || 'left'),
			bold  => (exists $opt->{boldleft} ? $opt->{boldleft} : 1),
			class => $opt->{leftclass},
			},
		right => {
			color => ($opt->{rightcolor} || $opt->{totalcolor}),
			align => ($opt->{rightalign} || 'right'),
			bold  => (exists $opt->{boldright} ? $opt->{boldright} : 1),
			class => $opt->{rightclass},
			},
		altrow => {
			color => ($opt->{altrowcolor} || $opt->{normalcolor} || $opt->{bgcolor}),
			align => ($opt->{normalalign} || 'left'),                      # Same as normal
			bold  => (exists $opt->{boldnormal} ? $opt->{boldnormal} : 0), # Same as normal
			class => ($opt->{altrowclass} || $opt->{normalclass}),
			},
		normal => {
			color => ($opt->{normalcolor} || $opt->{bgcolor}),
			align => ($opt->{normalalign} || 'left'),
			bold  => (exists $opt->{boldnormal} ? $opt->{boldnormal} : 0),
			class => $opt->{normalclass},
			},
		);
	
	# For each of the properties, create it's tag options hash
	foreach my $type ( values %properties ) {
		my $tag_options = {};
		$tag_options->{bgcolor} = $type->{color} if defined $type->{color};
		$tag_options->{align} = $type->{align} if defined $type->{align};
		$tag_options->{class} = $type->{class} if defined $type->{class};
		$tag_options->{nowrap} = undef if $opt->{nowrap};		
		$type->{tag_options} = $tag_options;
	}

	# Create the content buffer
	my @content = ();

	# Generate the inside rows/cells
	foreach my $row ( 0 .. $opt->{rows} ) {
		push @content, '  <tr>';
		foreach my $col ( 0 .. $opt->{columns} ) {
			# Establish the type
			my $type = $class->tableCellType( $row, $col, $opt );
			$type = $properties{$type};			
			
			# Get the cell contents and bold if needed
			my $cell_contents = $data->[$row]->[$col];
			$cell_contents = '&nbsp;' unless length $cell_contents;
			$cell_contents = "<b>$cell_contents</b>" if $type->{bold};
	
			# Create the cell tag and add it
			push @content, '    ' . AppLib::HTML->tagPair( 'td', $type->{tag_options}, $cell_contents );
		}
		push @content, '  </tr>';
	}

	return join "\n", @content;
}

# Wrap a bunch of rows in the double layer funky table
# Returns it as a single string, for convenience
sub tableWrap {
	my $class = shift;
	my $content = shift;
	my $opt = shift || {};
	
	# Set some defaults
	$opt->{cellpadding} = 3 unless exists $opt->{cellpadding};
	$opt->{innerwidth}  = 1 unless exists $opt->{innerwidth};
	$opt->{outerwidth}  = 1 unless exists $opt->{outerwidth};
	$opt->{bordercolor} = $defaultcolor->{border} unless exists $opt->{bordercolor};
	
	# Set the width of the table
	my $width = exists $opt->{width} ? " width='$opt->{width}'" : '';
	my $align = exists $opt->{align} ? " align='$opt->{align}'" : '';

	return qq~
	<table border='0' cellspacing='0' cellpadding='$opt->{outerwidth}'$width$align>
	<tr bgcolor='$opt->{bordercolor}'><td>
	<table border='0' cellspacing='$opt->{innerwidth}' cellpadding='$opt->{cellpadding}' width='100%'>
	$content
	</table></td></tr></table>
	~;	
}

sub tableCellType {
	my ( $class, $row, $col, $opt ) = @_;
	return 'top' if $row == 0 and $opt->{top};
	return 'left' if $col == 0 and $opt->{left};
	if ( $opt->{rows} and $opt->{bottom} and $row == $opt->{rows} ) {
		return 'bottom';
	}
	if ( $opt->{columns} and $opt->{right} and $col == $opt->{columns} ) {
		return 'right';
	}
	unless ( $row % 2 ) {
		return 'altrow';
	}
	return 'normal';
}

1;

__END__

# Copyright (C) 2000-2002 Adam Kennedy ( software.applib@ali.as )
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# Should you wish to utilise this software under a different licence, 
# please contact the author.

