package CVSMonitor::MetaData::Author;

# The CVSMonitor::MetaData::Author represents an Author. An Author can 
# exist within and can be derived from either a Module or a Repository. 

# An Author owns one or more ChangeSets, from which Files can be derived.

use strict;
use UNIVERSAL qw{isa can};




# An author is created on, and attached to, a source of ChangeSets.
# It pulls and maintains a list of ChangeSets from the ChangeSet source.
sub new {
	my $class = shift;
	my $name = shift or return undef;
	my $Source = shift or return undef;
	if ( can( $Source, 'getChangeSets' ) ) {
		# Get the changesets and filter to just out Author
		$Source = $Source->getChangeSets or return undef;
	}
	unless ( isa( $Source, 'ARRAY' ) 
		and 
		isa( $Source->[0], 'CVSMonitor::MetaData::ChangeSet' ) 
	) {
		return undef;
	}

	# Filter the changesets to our Author
	my @Filtered = grep { $_->author eq $name } @$Source;
	return undef unless @Filtered;
	
	# We have an array of ChangeSets, now create the object
	my $self = {
		name => $name,
		_CHANGESETS => \@Filtered,
		};
	return bless $self, $class;
}




#####################################################################
# Accessor methods

sub name { $_[0]->{name} }





#####################################################################
# Get derived objects

# Return a copy of the ChangeSets
sub getChangeSets {
	my $self = shift;
	return [ @{ $self->{_CHANGESETS} } ];
}

# Get the Repository we are in
sub Repository {
	my $self = shift;
	
	# Work our way up from the first ChangeSet
	my $ChangeSet = $self->{_CHANGESETS}->[0] or return undef;
	my $Module = $ChangeSet->Module or return undef;
	return $Module->Repository;
}

# Get the modules that we are in
sub getModules {
	my $self = shift;
	
	# Get the Repository we are in
	my $Repository = $self->Repository or return undef;
	
	# Get the modules, and filter down to the ones that contain our author
	my $Modules = $Repository->getModules or return undef;
	return grep { $_->containsAuthor( $self ) } @$Modules;
}

# Build a Range for the Author
sub getRange {
	my $self = shift;
	my $type = shift || 'Basics';
	my $arg = shift || 'month';
	my $RangeClass = 'CVSMonitor::Range::' 
		. ($type eq 'Files' ? 'Files' : 'Basics');

	# Do the entire thing to a particular depth
	if ( $arg eq 'year' or $arg eq 'month' ) {
		return $RangeClass->newFor( $self, $arg );
	}
	
	# Limit to a single year
	my $regex = CVSMonitor::DateDepth->seperate_regex( 'year' ) or return undef;
	if ( $arg =~ $regex ) {
		return $RangeClass->year( $arg );
	}
	
	# Limit to a single month
	$regex = CVSMonitor::DateDepth->seperate_regex( 'month' ) or return undef;
	if ( $arg =~ $regex ) {
		return $RangeClass->month( $arg );
	}
	
	return $self->andError( "Unsupported argument" );
}

# Get a populated Range for the Author
sub getRangePopulated {
	my $self = shift;
	my $type = shift || 'Basics';
	my $arg = shift;
	my $balanced = shift;
	
	# Get the empty Range
	my $Range = $self->getRange( $type, $arg ) or return undef;
	
	# Populate the range
	$Range->add( $self, $balanced ) or return undef;
	return $Range;
}		
	
1;

__END__

# Copyright (c) 2002 Adam Kennedy. All rights reserved.
# This program is free software; you can redistribute
# it and/or modify it under the same terms as Perl itself.
#
# The full text of the license can be found in the
# LICENSE.PERL file included with this module.
