package CVSMonitor::MetaData::Repository;

# The Repository allow the MetaData object to have a concept
# of repositories seperate from Modules.

# At the current time, it is a convenience.

use strict;
use UNIVERSAL 'isa';
use base 'CVSMonitor::MetaData::Abstract::Source';
use Class::Autouse qw{
	Validate::Net 
	Net::Ping
	Net::Telnet
	};

# Hold the default CVS port
use vars qw{$default_port};
BEGIN {
	$default_port = 2401;
}





#####################################################################
# Constructor and Friends

# In one sense, a repository is just an object form of a cvsroot,
# but with a lot more brains.

sub new {
	my $class = shift;
	my $name = shift or return undef;
	my $cvsroot = shift or return undef;
	my $options = shift || {};
	my $parent = shift;		
	$cvsroot =~ s/\s//g;
	
	# Create the basic object
	my $self = bless {
		_PARENT => undef, # All repositories need to be stored
		_MODULES => {},   # Repositories contain modules. 
		                  # All other object types are derived.
		type => undef,    # The type defines what other fields are available
		}, $class;

	# Check and set the symbolic name
	if ( CVSMonitor::MetaData->_checkFormat( 'symbol', $name ) ) {
		$self->{name} = $name;
	} else {
		return $self->andError( 'Symbolic name is invalid, should be 1-32 alhpanumeric characters only' );
	}
		
	# Make sure they arn't trying a local cvsroot
	unless ( $cvsroot =~ /\@/ ) {
		return $self->andError( "Local repositories are no longer permitted due to permission issues" );
	}

	# Initialise the cvsroot
	unless ( $self->_new_remote( $cvsroot, $options ) ) {
		return undef;
	}
		
	# Set the optional label and description
	if ( defined $options->{label} ) {
		$self->{label} = $options->{label};
	}
	if ( defined $options->{description} ) {
		$self->{description} = $options->{description};
	}
	
	# Check the parent if provided
	if ( $parent ) {
		unless ( isa( $parent, 'CVSMonitor::MetaData' ) ) {
			return $self->andError( "Parent argument to Repository constructor was not a MetaData object" );
		}
		$self->{_PARENT} = $parent;
	}
	
	return $self;
}

# Create a Repository object from a remote CVSROOT
sub _new_remote {
	my $self = shift;
	my $cvsroot = shift or return undef;
	my $options = shift || {};

	# Parse the cvsroot.
	# Split it first by the @ symbol.
	# The items after the @ holds the location of the repository.
	# The items before the @ are the access options.
	my ( $first, $second ) = split /\@/, $cvsroot;

	# Split both the access options and location by the colon
	my @access = split /\:/, $first;
	my @location = split /\:/, $second;
	push @access, '' if $first =~ /:$/;

	# Check the minimum number of elements in each
	return undef if scalar @access > 4;
	return undef if scalar @location > 3;

	# Process the access information.
	# The first element should be empty.
	return undef unless $access[0] eq '';
	shift @access;

	# The second element should be the type
	my $type = shift @access;
	if ( CVSMonitor::MetaData->_checkFormat( 'type', $type ) ) {
		$self->{type} = $type;
	} else {
		return undef;
	}

	# The third element should be the username
	my $username = shift @access;
	if ( CVSMonitor::MetaData->_checkFormat( 'username', $username ) ) {
		$self->{username} = $username;
	} else {
		return undef;
	}

	# IF the fourth element is defined, it's the password
	if ( scalar @access ) {
		my $password = shift @access;
		if ( length $password ) {
			if ( CVSMonitor::MetaData->_checkFormat( 'password', $password ) ) {
				$self->{password} = $password;
			} else {
				# No password
				$self->{password} = '';
			}
		}
	} else {
		# The password isn't in the CVSROOT, which means
		# they have to have password as an option.
		unless ( defined $options->{password} ) {
			return undef;
		} else {
			$self->{password} = $options->{password} || '';
		}
	}

	# Process the location information
	# The first element is the host
	my $host = shift @location;
	if ( Validate::Net->host( $host ) ) {
		$self->{host} = $host;
	} else {
		return undef;
	}

	# The LAST item should be the path
	my $path = shift @location;
	if ( CVSMonitor::MetaData->_checkFormat( 'path', $path ) ) {
		$self->{path} = $path;
	} else {
		return undef;
	}

	# If there's anything left, it's the port
	if ( scalar @location ) {
		my $port = shift @location;
		if ( Validate::Net->port( $port ) ) {
			$self->{port} = $port;
		} else {
			return undef;
		}
	}

	# Done, god that took a long time.
	# The object information should now be filled
	return $self;
}

# The Repository part of the MetaData ->reload method
sub _reload {
	my $self = shift;
	my $replacement = shift or return undef;
	
	# Work out which Modules are added, replaced, or removed
	my %state = ();
	foreach ( $replacement->getModuleNames ) {
		$state{$_} = 'add';
	}
	foreach ( $self->getModuleNames ) {
		$state{$_} = exists $state{$_} ? 'replace' : 'remove';
	}
	
	# Add/replace/remove the Modules
	foreach my $name ( keys %state ) {
		if ( $state{$name} eq 'add' ) {
			# Transfer the module from the replacement
			$self->{_MODULES}->{$name} = $replacement->{_MODULES}->{$name};
			$self->{_MODULES}->{$name}->{_PARENT} = $self;
			delete $replacement->{_MODULES}->{$name};
			
		} elsif ( $state{$name} eq 'replace' ) {
			# Hand off to the Module reloader
			$self->{_MODULES}->{$name}->_reload( $replacement->{_MODULES}->{$name} );
			
		} else {
			# Remove the module. Do this by emptying and reblessing as a
			# CVSMonitor::MetaData::Module::Deleted object, which just throws errors.
			# We do this in case something else is hanging on to a reference somewhere.
			$self->{_MODULES}->{$name}->_reloadDelete or return undef;
			delete $self->{_MODULES}->{$name};
		}
	}
	
	# Remove and replace all our scalar properties.
	foreach ( keys %$self ) {
		next if $_ eq '_MODULES';
		next if $_ eq '_PARENT';
		delete $self->{$_};
	}
	foreach ( keys %$replacement ) {
		next if $_ eq '_MODULES';
		next if $_ eq '_PARENT';
		$self->{$_} = $replacement->{$_};
	}
	delete $replacement->{_MODULES};
	
	# Done
	return 1;
}	
			
			
	


#####################################################################
# Methods on a valid LOOKING repository

sub name     { $_[0]->{name} }
sub type     { $_[0]->{type} }
sub username { $_[0]->{username} }
sub password { $_[0]->{password} }
sub host     { $_[0]->{host} }
sub port     { $_[0]->{port} || $default_port }
sub path     { $_[0]->{path} }
sub MetaData { $_[0]->{_PARENT} }

# Get and set the label
sub getLabel {
	my $self = shift;
	return $self->{label} || $self->cvsroot;
}
sub setLabel {
	my $self = shift;
	$self->{label} = shift;
	return 1;
}

# Get and set the description
sub getDescription {
	my $self = shift;
	return $self->{description} || "A CVS Repository";
}
sub setDescription {
	my $self = shift;
	$self->{description} = shift;
	return 1;
}

# Is the repository valid
sub exists {
	my $self = shift or return undef;
	
	# Try to connect to the port using telnet
	my $port = $self->port;
	my $host = Net::Telnet->new() or return undef;
	my $rv;
	eval { 
		$rv = $host->open( 
			Host => $self->{host}, 
			Port => $port,
			Timeout => 5
			);
	};
	return $self->andError( $@ ) if $@;
	
	# Handle errors
	unless ( $rv ) {
		return $self->andError( "Host exists but nothing listening on port $port" );
	}
	
	return 1;
}	

# Test a connection to a repository
sub test {
	my $self = shift;

	# Prepare by establishing some values
	my $cvsroot = $self->cvsroot;
	my $dir = Cwd::fastcwd();
	
	# Do the test
	chdir $self->MetaData->root or return undef;
	my $rv = system( "cvs -d $cvsroot -Q checkout -c" );
	chdir $dir or return undef;
	
	return $rv ? 0 : 1;
}

# Recover the full cvsroot
sub cvsroot {
	my $self = shift;
	
	# Do this the long way, to handle ports and shit
	my $first = ':' . ($self->{type}||'')
		. ':' . ($self->{username}||'')
		. ':' . ($self->{password}||'');
	my $second = $self->{host}
		. ($self->{port} ? (':' . $self->{port}) : '')
		. ':' . ($self->{path}||'');
	return "$first\@$second";
}

# Public version of the cvsroot, just don't show the password
sub publicCvsroot {
	my $self = shift;

	# Do this the long way, to handle ports and shit
	my $first = ':' . ($self->{type}||'')
		. ':' . ($self->{username}||'');
	my $second = $self->{host}
		. ($self->{port} ? (':' . $self->{port}) : '')
		. ':' . ($self->{path}||'');
	return "$first\@$second";
}





#####################################################################
# Working with modules

# Get the module names.
# Return in sorted order.
sub getModuleNames { sort keys %{ $_[0]->{_MODULES} } }

# Get the Modules.
# Return in sorted order.
sub getModules { 
	my $self = shift;
	return map { $self->{_MODULES}->{$_} }
		sort keys %{ $self->{_MODULES} };
}

# Get a given module by name
sub getModule { $_[0]->{_MODULES}->{$_[1]} }

# Add a module to a Repository
sub addModule {
	my $self = shift;
	my $name = shift or return undef;
	my $path = shift or return undef;
	
	# Does something by that name already exist?
	if ( exists $self->{_MODULES}->{$name} ) {
		return $self->andError( "A module by the name '$name' already exists" );
	}
	
	# Do we hold a lock?
	unless ( $self->MetaData->haveLock ) {
		return $self->andError( "Cannot create module. We do not have a lock on the index file" );
	}	
	
	# Create the module object
	my $Module = CVSMonitor::MetaData::Module->new( $name, $path, $self );
	return undef unless defined $Module;
	
	# Add it
	$self->{_MODULES}->{$name} = $Module;
	$self->dirty;
	return 1;
}

# Delete ( outright ) a module
sub deleteModule {
	my $self = shift;
	my $name = shift or return undef;
	
	# Get the module.
	my $Module = $self->getModule( $name );
	return $self->andError( "Module '$name' does not exist" ) unless $Module;

	# Is the module locked?
	if ( $Module->locked ) {
		return $self->andError( "Cannot delete module. The module is locked" );
	}

	# Do we hold a metadata lock?
	unless ( $self->MetaData->haveLock ) {
		return $self->andError( "Cannot delete module. We do not have a lock on the index file" );
	}	
	
	# Delete it's files
	unless ( $Module->purge ) {
		return $self->andError( "Error removing files for the module" );
	}
	
	# Remove it's entry
	delete $self->{_MODULES}->{$name}->{_PARENT};
	delete $self->{_MODULES}->{$name};
	$self->dirty;
	
	return 1;
}

# Update all updatable modules
sub update {
	my $self = shift;
	foreach my $Module ( $self->getModules ) {
		unless ( $Module->update ) {
			my $name = $Module->fullName;
			return $self->andError( "Error while updating module '$name'" );
		}
	}	
	return 1;
}

# Get the update cost. Add together the costs for all the module in the repository
sub updateCost {
	my $self = shift;
	my $total = 0;
	
	# Iterate over the modules and add them up
	foreach my $Module ( $self->getModules ) {
		my $cost = $Module->updateCost;
		return undef unless defined $cost;
		$total += $cost;
	}
	
	return $total;
}

# Conveniece method.
# Pass through to our MetaData
sub dirty { $_[0]->MetaData->dirty }






#####################################################################
# Getting other things

# Are ALL the modules within the Repository available?
sub available {
	my $self = shift;
	foreach my $Module ( $self->getModules ) {
		return 0 unless $Module->available;
	}
	return 1;
}
	
# Get an Activity object for this
sub getActivity { CVSMonitor::MetaData::Activity->new( shift ) }


# Build a Range for the Repository
sub getRange {
	my $self = shift;
	my $type = shift || 'Basics';
	my $arg = shift || 'month';
		my $RangeClass = 'CVSMonitor::Range::' 
		. ($type eq 'Files' ? 'Files' : 'Basics');
	
	# Do the entire thing to a particular depth
	if ( $arg eq 'year' or $arg eq 'month' ) {
		return $RangeClass->newFor( $self, $arg );
	}
	
	# Limit to a single year
	my $regex = CVSMonitor::DateDepth->seperate_regex( 'year' ) or return undef;
	if ( $arg =~ $regex ) {
		return $RangeClass->year( $arg );
	}
	
	# Limit to a single month
	$regex = CVSMonitor::DateDepth->seperate_regex( 'month' ) or return undef;
	if ( $arg =~ $regex ) {
		return $RangeClass->month( $arg );
	}
	
	return $self->andError( "Unsupported argument" );
}

# Build a populated Range from the Module
sub getRangePopulated {
	my $self = shift;
	my $type = shift || 'Basics';
	my $arg = shift;
	
	# Get the empty Range
	my $Range = $self->getRange( $type, $arg ) or return undef;
	
	# Populate the range
	$Range->add( $self, @_ ) or return undef;
	return $Range;
}

# Get the merged changesets for ALL modules.
# Optionally within a date period
sub getChangeSets {
	my $self = shift;
	my $date = shift;
	my @Merged = ();
	
	# Get the list of modules
	foreach my $Module ( $self->getModules ) {
		# Get it's changesets
		my $ChangeSets = $Module->getChangeSets( $date );
		next unless $ChangeSets;
		push @Merged, @$ChangeSets;
	}
	
	return scalar @Merged ? \@Merged : 0;
}

# Return the names of all the modules the repository says it has
sub getAdvertisedModuleNames {
	my $self = shift;
	
	# Prepare by establishing some values
	my $cvsroot = $self->cvsroot;
	my $dir = Cwd::fastcwd();
	
	# Do the test
	chdir $self->MetaData->root or return undef;
	my @output = `cvs -d $cvsroot -Q checkout -c`;
	chdir $dir or return undef;

	# Process the output
	chomp @output;
	return @output;
}	





#####################################################################
# Loading and Saving

# Generate a freezable version of the Repository
# Create a stub to freeze into the index
sub _freezable {
        my $self = shift;
        
        # Copy the main hash, and remove unwanted things
        my $copy = bless { %{ $self } }, ref $self;
        delete $copy->{_PARENT};

	# Copy the modules hash, and replace with module
	# freeze stubs	
	my $modules = { %{ $copy->{_MODULES} } };
	foreach ( keys %$modules ) {
		$modules->{$_} = $modules->{$_}->_freezable;
		$modules->{$_}->{_PARENT} = $copy;
	}
	$copy->{_MODULES} = $modules;	

        return $copy;
}

	
	




#####################################################################
# Utility Methods

sub andError { $CVSMonitor::MetaData::errstr = $_[1]; undef }

1;

__END__

# Copyright (c) 2002 Adam Kennedy. All rights reserved.
# This program is free software; you can redistribute
# it and/or modify it under the same terms as Perl itself.
#
# The full text of the license can be found in the
# LICENSE.PERL file included with this module.
